# -*- coding: utf-8 -*-

# FLEDGE_BEGIN
# See: http://fledge-iot.readthedocs.io/
# FLEDGE_END

import copy
from datetime import datetime
from enum import IntEnum
from aiohttp import web
import json

from fledge.common.storage_client.payload_builder import PayloadBuilder
from fledge.common.storage_client.exceptions import StorageServerError
from fledge.services.core import connect
from fledge.common.audit_logger import AuditLogger
from fledge.common import logger

__author__ = "Amarendra K. Sinha, Ashish Jabble, Massimiliano Pinto"
__copyright__ = "Copyright (c) 2017-2018 OSIsoft, LLC"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

__DEFAULT_LIMIT = 20
__DEFAULT_OFFSET = 0

_help = """
    -------------------------------------------------------------------------------
    | GET POST        | /fledge/audit                                            |
    | GET             | /fledge/audit/logcode                                    |
    | GET             | /fledge/audit/severity                                   |
    -------------------------------------------------------------------------------
"""

_logger = logger.setup(__name__)


class Severity(IntEnum):
    """ Enumeration for log.severity """
    # TODO: FOGL-1100, no info for 3
    SUCCESS = 0
    FAILURE = 1
    WARNING = 2
    INFORMATION = 4

####################################
#  Audit Trail
####################################


async def create_audit_entry(request):
    """ Creates a new Audit entry

    Args:
        request: POST /fledge/audit

        {
          "source"   : "LMTR", # 5 char max
          "severity" : "WARNING",
          "details"  : {
                        "message" : "Engine oil pressure low"
                      }
        }

    :Example:

        curl -X POST -d '{"source":"LMTR","severity":"WARNING","details":{"message":"Engine oil pressure low"}}
        http://localhost:8081/fledge/audit

    Returns:
        json object representation of created audit entry

        {
          "timestamp" : "2017-06-21T09:39:51.8949395",
          "source"    : "LMTR",
          "severity"  : "WARNING",
          "details"   : {
                         "message" : "Engine oil pressure low"
                        }
        }
    """
    return_error = False
    err_msg = "Missing required parameter"

    payload = await request.json()

    severity = payload.get("severity")
    source = payload.get("source")
    details = payload.get("details")

    if severity is None or severity == "":
        err_msg += " severity"
        return_error = True
    if source is None or source == "":
        err_msg += " source"
        return_error = True
    if details is None:
        err_msg += " details"
        return_error = True

    if return_error:
        raise web.HTTPBadRequest(reason=err_msg)

    if not isinstance(details, dict):
        raise web.HTTPBadRequest(reason="Details should be a valid json object")

    try:
        audit = AuditLogger()
        await getattr(audit, str(severity).lower())(source, details)

        # Set timestamp for return message
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

        message = {'timestamp': str(timestamp),
                   'source': source,
                   'severity': severity,
                   'details': details
                   }
    except AttributeError as e:
        # Return error for wrong severity method
        err_msg = "severity type {} is not supported".format(severity)
        _logger.error("Error in create_audit_entry(): %s | %s", err_msg, str(e))
        raise web.HTTPNotFound(reason=err_msg, body=json.dumps({"message": err_msg}))
    except StorageServerError as ex:
        if int(ex.code) in range(400, 500):
            err_msg = 'Audit entry cannot be logged. {}'.format(ex.error['message'])
            raise web.HTTPBadRequest(body=json.dumps({"message": err_msg}))
        else:
            err_msg = 'Failed to log audit entry. {}'.format(ex.error['message'])
            raise web.HTTPInternalServerError(body=json.dumps({"message": err_msg}))
    except Exception as ex:
        raise web.HTTPInternalServerError(reason=str(ex), body=json.dumps({"message": str(ex)}))
    else:
        return web.json_response(message)


async def get_audit_entries(request):
    """ Returns a list of audit trail entries sorted with most recent first and total count
        (including the criteria search if applied)

    :Example:

        curl -X GET http://localhost:8081/fledge/audit

        curl -X GET http://localhost:8081/fledge/audit?limit=5

        curl -X GET "http://localhost:8081/fledge/audit?limit=5&skip=3"

        curl -X GET http://localhost:8081/fledge/audit?skip=2

        curl -X GET http://localhost:8081/fledge/audit?source=PURGE

        curl -X GET http://localhost:8081/fledge/audit?source=NTFSD,NTFSN,NTFAD,NTFST,NTFDL,NTFCL
        curl -X GET http://localhost:8081/fledge/audit?severity=FAILURE

        curl -X GET "http://localhost:8081/fledge/audit?source=LOGGN&severity=INFORMATION&limit=10"

        curl -X GET "http://localhost:8081/fledge/audit?source=CONAD&since=2022-10-10%2009:31:32"
    """

    limit = __DEFAULT_LIMIT
    if 'limit' in request.query and request.query['limit'] != '':
        try:
            limit = int(request.query['limit'])
            if limit < 0:
                raise ValueError
        except ValueError:
            raise web.HTTPBadRequest(reason="Limit must be a positive integer")

    offset = __DEFAULT_OFFSET
    if 'skip' in request.query and request.query['skip'] != '':
        try:
            offset = int(request.query['skip'])
            if offset < 0:
                raise ValueError
        except ValueError:
            raise web.HTTPBadRequest(reason="Skip/Offset must be a positive integer")

    # If microsend is required then add .%f into the __DATE_FORMAT & remove the split from datetime string conversion
    __DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
    if 'since' in request.query and request.query['since'] != '':
        try:
            datetime.strptime(request.query['since'], __DATE_FORMAT)
        except ValueError:
            msg = "Incorrect date format, should be {}".format(__DATE_FORMAT)
            raise web.HTTPBadRequest(reason=msg, body=json.dumps({"message": msg}))

    source = None
    source_list = []
    if 'source' in request.query and request.query['source'] != '':
        try:
            source = request.query.get('source')
            source_list = source.split(',')
            # SELECT * FROM log_codes
            storage_client = connect.get_storage_async()
            result = await storage_client.query_tbl("log_codes")
            log_codes = [key['code'] for key in result['rows']]
            for code in source_list:
                if code not in log_codes:
                    raise ValueError(code)
        except ValueError as e:
            raise web.HTTPBadRequest(reason="{} is not a valid source".format(str(e)))

    severity = None
    if 'severity' in request.query and request.query['severity'] != '':
        try:
            severity = Severity[request.query['severity'].upper()].value
        except KeyError as ex:
            raise web.HTTPBadRequest(reason="{} is not a valid severity".format(ex))

    try:
        # HACK: This way when we can more future we do not get an exponential
        # explosion of if statements
        payload = PayloadBuilder().SELECT("code", "level", "log", "ts")\
            .ALIAS("return", ("ts", 'timestamp')).FORMAT("return", ("ts", "YYYY-MM-DD HH24:MI:SS.MS"))\
            .WHERE(['1', '=', 1])

        if source is not None:
            if len(source_list) == 1:
                payload.AND_WHERE(['code', '=', source])
            else:
                payload = PayloadBuilder().SELECT("code", "level", "log", "ts") \
                    .ALIAS("return", ("ts", 'timestamp')).FORMAT("return", ("ts", "YYYY-MM-DD HH24:MI:SS.MS"))
                payload.WHERE(['code', 'in', source_list])
        if severity is not None:
            payload.AND_WHERE(['level', '=', severity])

        _and_where_payload = payload.chain_payload()
        # SELECT *, count(*) OVER() FROM log - No support yet from storage layer
        # TODO: FOGL-740, FOGL-663 once ^^ resolved we should replace below storage call for getting total rows
        _and_where_copy = copy.deepcopy(_and_where_payload)
        total_count_payload = PayloadBuilder(_and_where_copy).AGGREGATE(["count", "*"])\
            .ALIAS("aggregate", ("*", "count", "count")).payload()

        # SELECT count (*) FROM log <_and_where_payload>
        storage_client = connect.get_storage_async()
        result = await storage_client.query_tbl_with_payload('log', total_count_payload)
        total_count = result['rows'][0]['count']

        payload = PayloadBuilder(_and_where_payload)
        payload.ORDER_BY(['ts', 'desc'])
        payload.LIMIT(limit)

        if offset > 0:
            payload.OFFSET(offset)

        # SELECT * FROM log <payload.payload()>
        results = await storage_client.query_tbl_with_payload('log', payload.payload())
        rows = results['rows']
        # If 'since' datetime string param is passed then filter the records internally from the actual storage result
        if 'since' in request.query:
            since_dt = datetime.strptime(request.query['since'].split('.', 1)[0], __DATE_FORMAT)
            temp_rows = []
            for row in results["rows"]:
                convert_dt = datetime.strptime(row['timestamp'].split('.', 1)[0], __DATE_FORMAT)
                if since_dt <= convert_dt:
                    temp_rows.append(row)
            rows = temp_rows
            total_count = len(rows)
        res = []
        for row in rows:
            r = dict()
            r["details"] = row["log"]
            severity_level = int(row["level"])
            r["severity"] = Severity(severity_level).name if severity_level in (0, 1, 2, 4) else "UNKNOWN"
            r["source"] = row["code"]
            r["timestamp"] = row["timestamp"]
            res.append(r)
    except Exception as ex:
        raise web.HTTPInternalServerError(reason=str(ex))
    else:
        return web.json_response({'audit': res, 'totalCount': total_count})


async def get_audit_log_codes(request):
    """
    Args:
        request:

    Returns:
           an array of log codes with description

    :Example:

        curl -X GET http://localhost:8081/fledge/audit/logcode
    """
    storage_client = connect.get_storage_async()
    result = await storage_client.query_tbl('log_codes')

    return web.json_response({'logCode': result['rows']})


async def get_audit_log_severity(request):
    """
    Args:
        request:

    Returns:
            an array of audit severity enumeration key index values

    :Example:

        curl -X GET http://localhost:8081/fledge/audit/severity
    """
    results = []
    for _severity in Severity:
        data = {'index': _severity.value, 'name': _severity.name}
        results.append(data)

    return web.json_response({"logSeverity": results})
